from typing import Literal, List
import urllib
import json
import xml.etree.ElementTree as ET

from ..structs.landmark import Landmark

ElementTypes = List[Literal['way', 'node', 'relation']]



def OverpassQueryBuilder(area: tuple, element_types: ElementTypes, selector: str, 
                         conditions=[], out='center'):
    """
    Constructs a query string for the Overpass API to retrieve OpenStreetMap (OSM) data.

    Args:
        area (tuple): A tuple representing the geographical search area, typically in the format 
                      (radius, latitude, longitude). The first element is a string like "around:2000" 
                      specifying the search radius, and the second and third elements represent 
                      the latitude and longitude as floats or strings.
        element_types (list[str]): A list of OSM element types to search for. Must be one or more of 
                                   'Way', 'Node', or 'Relation'.
        selector (str): The key or tag to filter the OSM elements (e.g., 'amenity', 'highway', etc.).
        conditions (list, optional): A list of conditions to apply as additional filters for the 
                                     selected OSM elements. The conditions should be written in 
                                     the Overpass QL format, and they are combined with '&&' if 
                                     multiple are provided. Defaults to an empty list.
        out (str, optional): Specifies the output type, such as 'center', 'body', or 'tags'. 
                             Defaults to 'center'.

    Returns:
        str: The constructed Overpass QL query string.

    Notes:
        - If no conditions are provided, the query will just use the `selector` to filter the OSM 
          elements without additional constraints.
        - The search area must always formatted as "(radius, lat, lon)".
    """
    if not isinstance(conditions, list) :
        conditions = [conditions]

    query = '('
    search_area = f"({', '.join(map(str, area))})"

    if conditions :
        conditions = '(if: ' + ' && '.join(conditions) + ')'
    else :
        conditions = ''

    for elem in element_types :
        query += elem + '[' + selector + ']' + conditions + search_area + ';'

    query += ');' + f'out {out};'

    return query


def send_overpass_query(query: str) -> dict:
    """
    Sends the Overpass QL query to the Overpass API and returns the parsed JSON response.

    Args:
        query (str): The Overpass QL query to be sent to the Overpass API.

    Returns:
        dict: The parsed JSON response from the Overpass API, or None if the request fails.
    """

    # Define the Overpass API endpoint
    overpass_url = "https://overpass-api.de/api/interpreter"

    # Prepare the data to be sent as POST request, encoded as bytes
    data = urllib.parse.urlencode({'data': query}).encode('utf-8')

    # Create a custom header with a User-Agent
    headers = {
        'User-Agent': 'Mozilla/5.0 (compatible; OverpassQuery/1.0; +http://example.com)',
    }

    try:
        # Create a Request object with the specified URL, data, and headers
        request = urllib.request.Request(overpass_url, data=data, headers=headers)
        
        # Send the request and read the response
        with urllib.request.urlopen(request) as response:
            # Read and decode the response
            response_data = response.read().decode('utf-8')
            return ET.fromstring(response_data)

    except urllib.error.URLError as e:
        print(f"Error connecting to Overpass API: {e}")
        return None
    except json.JSONDecodeError:
        print("Error decoding the JSON response from Overpass API.")
        return None


def parse_result(root: ET.Element, elem_type) -> List[Landmark]:

    landmarks = []
    if root is None :
        return landmarks

    for osm_type in ['node', 'way', 'relation'] :
        for elem in root.findall(osm_type):

            # Extract basic info from the landmark.
            name = elem.find("tag[@k='name']").get('v') if elem.find("tag[@k='name']") is not None else None
            center = elem.find('center')
            tags = elem.findall('tag')

            # Extract the center latitude and longitude if available.
            if name is not None and center is not None:
                lat = float(center.get('lat'))
                lon = float(center.get('lon'))
                coords = tuple((lat, lon))
            else :
                continue
            
            # Convert this to Landmark object
            landmark = Landmark(name=name,
                                type=elem_type,
                                location=coords,
                                osm_id=elem.get('id'), 
                                osm_type=osm_type,
                                attractiveness=0,
                                n_tags=len(tags))

            # Browse through tags to add information to landmark.
            for tag in tags:
                key = tag.get('k')
                value = tag.get('v')

                # Skip this landmark if not suitable.
                if key == 'building:part' and value == 'yes' :
                    break
                if 'disused:' in key :
                    break
                if 'boundary:' in key :
                    break
                if 'shop' in key and elem_type != 'shopping' :
                    break
                # if value == 'apartments' :
                #     break
                
                # Fill in the other attributes.
                if key == 'image' :
                    landmark.image_url = value
                if key == 'website' :
                    landmark.website_url = value
                if key == 'place_of_worship' :
                    landmark.is_place_of_worship = True
                if key == 'wikipedia' :
                    landmark.wiki_url = value
                if key == 'name:en' :
                    landmark.name_en = value
                if 'building:' in key or 'pay' in key :
                    landmark.n_tags -= 1
                
                # Set the duration.
                if value in ['museum', 'aquarium', 'planetarium'] :
                    landmark.duration = 60
                elif value == 'viewpoint' :
                    landmark.is_viewpoint = True
                    landmark.duration = 10
                elif value == 'cathedral' :
                    landmark.is_place_of_worship = False
                    landmark.duration = 10
                else :
                    landmark.duration = 5

            else: 
                set_score(landmark, elem_type)
                landmarks.append(landmark)
            continue

    return landmarks



def set_score(landmark: Landmark, landmarktype: str) :

        score = landmark.n_tags**1.15
        if landmark.wiki_url :
            score *= 1.1
        if landmark.image_url :
            score *= 1.1
        if landmark.website_url :
            score *= 1.1
        if landmark.is_place_of_worship :
            score *= 0.65
        if landmark.is_viewpoint :
            # print(f"{landmark.name}:  n_tags={landmark.n_tags} and score={score*3*1.35*10}")
            score *= 3
        if landmarktype == 'nature' :
            score *= 1.35

        landmark.attractiveness = int(score * 10)
